﻿"use strict";

function Complex(x, y) {
	this.x = x;
	this.y = y;
}

function complex(x, y) {
	if (y === undefined)
		y = 0;
		
	return new Complex(x, y);
}

Complex.zero = new Complex(0, 0);
Complex.one = new Complex(1, 0);

Complex.polar = function(modulus, angle) {
	return new Complex(modulus * Math.cos(angle), modulus * Math.sin(angle));
};

Complex.abs = function(value) {
	if (typeof value === 'number')
		return Math.abs(value);

	return value.abs();
};

Complex.arg = function(value) {
	if (typeof value === 'number')
		return Math.atan2(0, value);

	return value.arg();
};

Complex.add = function(a, b) {
	if ((typeof a === 'number') && (typeof b === 'number'))
		return a + b;
	
	if (typeof a === 'number')
		a = new Complex(a, 0);

	return a.add(b);
};

Complex.sub = function(a, b) {
	if ((typeof a === 'number') && (typeof b === 'number'))
		return a - b;
	
	if (typeof a === 'number')
		a = new Complex(a, 0);

	return a.sub(b);
};

Complex.mul = function(a, b) {
	if ((typeof a === 'number') && (typeof b === 'number'))
		return a * b;
	
	if (typeof a === 'number')
		a = new Complex(a, 0);

	return a.mul(b);
};

Complex.div = function(a, b) {
	if ((typeof a === 'number') && (typeof b === 'number'))
		return a / b;
	
	if (typeof a === 'number')
		a = new Complex(a, 0);

	return a.div(b);
};

Complex.sqrt = function(value) {
	if (typeof value === 'number')
		return value >= 0 ? Math.sqrt(value) : (new Complex(value, 0)).sqrt();

	return value.sqrt();
};

Complex.prototype.equals = function(value) {
	if (typeof value === 'number')
		return this.y === 0 && this.x === value;
		
	return this.x === value.x && this.y === value.y;
};

Complex.prototype.abs = function() {
	return Math.sqrt(this.x * this.x + this.y * this.y);
};

Complex.prototype.length = Complex.prototype.abs;

Complex.prototype.abs2 = function() {
	return this.x * this.x + this.y * this.y;
};

Complex.prototype.angle = function() {
	return Math.atan2(this.y, this.x);
};

Complex.prototype.arg = Complex.prototype.angle;

Complex.prototype.cc = function() {
	return new Complex(this.x, -this.y);
};

Complex.prototype.rotate = function(angle) {
	var sin = Math.sin(angle);
	var cos = Math.cos(angle);
	
	return new Complex(
		this.x * cos - this.y * sin,
		this.x * sin + this.y * cos
	);
};

Complex.prototype.add = function(value) {
	if (typeof value === 'number')
		return new Complex(this.x + value, this.y)

	return new Complex(this.x + value.x, this.y + value.y);
};

Complex.prototype.sub = function(value) {
	if (typeof value === 'number')
		return new Complex(this.x - value, this.y)

	return new Complex(this.x - value.x, this.y - value.y);
};

Complex.prototype.mul = function(value) {
	if (typeof value === 'number')
		return new Complex(value * this.x, value * this.y);
		
	return new Complex(
		this.x * value.x - this.y * value.y,
		this.x * value.y + this.y * value.x
	);
};

Complex.prototype.div = function(value) {
	if (typeof value === 'number')
		return new Complex(this.x / value, this.y / value);

	var tmp = value.x * value.x + value.y * value.y;

	return new Complex(
		(this.x * value.x + this.y * value.y) / tmp,
		(this.y * value.x - this.x * value.y) / tmp
	);
};

Complex.prototype.invert = function() {
	var tmp = this.x * this.x + this.y * this.y;
	
	return new Complex(this.x / tmp, -this.y / tmp);
};

Complex.prototype.exp = function() {
	return Complex.polar(Math.exp(this.x), this.y);
};

Complex.prototype.log = function() {
	return new Complex(Math.log(this.abs()), this.arg());
};

Complex.prototype.power = function(value) {
	return this.equals(Complex.zero) ? Complex.zero : this.log().mul(value).exp();
};

Complex.prototype.sqrt = function() {
	//return this.power(0.5);
	return Complex.polar(
		Math.sqrt(this.abs()),
		0.5 * this.arg()
	);
};

Complex.prototype.sin = function() {
	var z1 = new Complex(-this.y, this.x);
	var z2 = new Complex(this.y, -this.x);
	return z1.exp().sub(z2.exp()).mul(new Complex(0, -0.5));
};

Complex.prototype.cos = function() {
	var z1 = new Complex(-this.y, this.x);
	var z2 = new Complex(this.y, -this.x);
	return z1.exp().add(z2.exp()).mul(0.5);
};

Complex.prototype.toString = function() {
	return '(' + this.x.toString() + ', ' + this.y.toString() + ')';
};
